@@ -87,12 +87,12 @@ garcon_menu_separator_finalize (GObject *object)
 
 
 /**
- * garcon_menu_separator_get_default:
+ * garcon_menu_separator_get_default: (constructor)
  *
- * Returns the default #GarconMenuSeparator.
- *
- * Return value: the default #GarconMenuSeparator. The returned object
+ * Returns the default #GarconMenuSeparator. The returned object
  * should be unreffed with g_object_unref() when no longer needed.
+ *
+ * Returns: (transfer full): the default #GarconMenuSeparator.
  */
 GarconMenuSeparator*
 garcon_menu_separator_get_default (void)
@@ -103,7 +103,7 @@ garcon_menu_separator_get_default (void)
     {
       /* Create a new separator */
       separator = g_object_new (GARCON_TYPE_MENU_SEPARATOR, NULL);
-      g_object_add_weak_pointer (G_OBJECT (separator), (gpointer) &separator);
+      g_object_add_weak_pointer (G_OBJECT (separator), (gpointer *) &separator);
     }
   else
     {
