@@ -215,6 +215,9 @@ struct _GarconMenuPrivate
 
   /* idle reload-required to group events */
   guint                idle_reload_required_id;
+
+  /* asynchronous load from GarconGtkMenu */
+  GMutex               load_lock;
 };
 
 
@@ -328,6 +331,8 @@ garcon_menu_init (GarconMenu *menu)
   menu->priv->pool = garcon_menu_item_pool_new ();
   menu->priv->uses_custom_path = TRUE;
   menu->priv->changed_files = NULL;
+  menu->priv->file_changed_idle = 0;
+  g_mutex_init (&menu->priv->load_lock);
   menu->priv->idle_reload_required_id = 0;
 
   /* Take reference on the menu item cache */
@@ -391,6 +396,7 @@ garcon_menu_finalize (GObject *object)
 
   /* Clear resources allocated in the load process */
   garcon_menu_clear (menu);
+  g_mutex_clear (&menu->priv->load_lock);
 
   /* Free file */
   if (menu->priv->file != NULL)
@@ -591,14 +597,11 @@ garcon_menu_get_name (GarconMenu *menu)
  * Returns the #GarconMenuDirectory of @menu or %NULL if @menu has
  * no valid directory element.
  *
- * The returned object should be unreffed with g_object_unref() when
- * no longer needed.
- *
  * The menu directory may contain a lot of useful information about
  * the menu like the display and icon name, desktop environments it
  * should show up in etc.
  *
- * Returns: (transfer full) (nullable): a #GarconMenuDirectory
+ * Returns: (transfer none) (nullable): a #GarconMenuDirectory
  */
 GarconMenuDirectory*
 garcon_menu_get_directory (GarconMenu *menu)
@@ -674,6 +677,8 @@ garcon_menu_load (GarconMenu   *menu,
   g_return_val_if_fail (GARCON_IS_MENU (menu), FALSE);
   g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
+  g_mutex_lock (&menu->priv->load_lock);
+
   /* Make sure to reset the menu to a loadable state */
   garcon_menu_clear (menu);
 
@@ -690,7 +695,7 @@ garcon_menu_load (GarconMenu   *menu,
       /* Build the ${XDG_MENU_PREFIX}applications.menu filename */
       prefix = g_getenv ("XDG_MENU_PREFIX");
       relative_filename = g_strconcat ("menus", G_DIR_SEPARATOR_S,
-                                       prefix != NULL ? prefix : GARCON_DEFAULT_MENU_PREFIX,
+                                       prefix != NULL ? prefix : _GARCON_DEFAULT_MENU_PREFIX,
                                        "applications.menu", NULL);
 
       /* Search for the menu file in user and system config dirs */
@@ -706,6 +711,7 @@ garcon_menu_load (GarconMenu   *menu,
           g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_NOENT,
                        _("File \"%s\" not found"), relative_filename);
           g_free (relative_filename);
+          g_mutex_unlock (&menu->priv->load_lock);
           return FALSE;
         }
 
@@ -742,7 +748,10 @@ garcon_menu_load (GarconMenu   *menu,
   g_object_unref (parser);
 
   if (!success)
-    return FALSE;
+    {
+      g_mutex_unlock (&menu->priv->load_lock);
+      return FALSE;
+    }
 
   /* Generate submenus */
   garcon_menu_resolve_menus (menu);
@@ -752,7 +761,10 @@ garcon_menu_load (GarconMenu   *menu,
 
   /* Abort if the cancellable was cancelled */
   if (g_cancellable_set_error_if_cancelled (cancellable, error))
-    return FALSE;
+    {
+      g_mutex_unlock (&menu->priv->load_lock);
+      return FALSE;
+    }
 
   desktop_id_table = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
 
@@ -769,6 +781,8 @@ garcon_menu_load (GarconMenu   *menu,
   /* Initiate file system monitoring */
   garcon_menu_start_monitoring (menu);
 
+  g_mutex_unlock (&menu->priv->load_lock);
+
   return TRUE;
 }
 
@@ -781,7 +795,7 @@ garcon_menu_load (GarconMenu   *menu,
  * Returns a sorted list of #GarconMenu submenus of @menu. The list
  * should be freed with g_list_free().
  *
- * Returns: (transfer full) (element-type GarconMenu): a sorted list
+ * Returns: (transfer container) (element-type GarconMenu): a sorted list
  * of #GarconMenu.
  */
 GList *
@@ -835,7 +849,7 @@ garcon_menu_add_menu (GarconMenu *menu,
  *
  * Looks in @menu for a submenu with @name as name.
  *
- * Returns: (transfer full) (nullable): a #GarconMenu or %NULL.
+ * Returns: (transfer none) (nullable): a #GarconMenu or %NULL.
  */
 GarconMenu *
 garcon_menu_get_menu_with_name (GarconMenu  *menu,
@@ -866,7 +880,7 @@ garcon_menu_get_menu_with_name (GarconMenu  *menu,
  *
  * Returns the parent #GarconMenu or @menu.
  *
- * Returns: (transfer full) (nullable): a #GarconMenu or %NULL
+ * Returns: (transfer none) (nullable): a #GarconMenu or %NULL
  * if @menu is the root menu.
  */
 GarconMenu *
@@ -1364,7 +1378,7 @@ garcon_menu_remove_deleted_menus (GarconMenu *menu)
  * Get the item pool of the menu. This pool contains all items in this
  * menu (for that of its submenus).
  *
- * Returns: (transfer full): a #GarconMenuItemPool.
+ * Returns: (transfer none): a #GarconMenuItemPool.
  */
 GarconMenuItemPool*
 garcon_menu_get_item_pool (GarconMenu *menu)
@@ -1396,7 +1410,7 @@ items_collect (const gchar    *desktop_id,
  * The caller is responsible to free the returned list using
  * g_list_free() when no longer needed.
  *
- * Returns: (element-type GarconMenuItem) (transfer full): list
+ * Returns: (element-type GarconMenuItem) (transfer container): list
  * of #GarconMenuItem included in @menu.
  */
 GList *
@@ -1525,7 +1539,7 @@ layout_elements_collect (GList **dest_list,
  * Returns a list of #GarconMenuItem or %NULL. Free the list with
  * g_list_free().
  *
- * Returns: (element-type GarconMenuItem) (nullable) (transfer full):
+ * Returns: (element-type GarconMenuItem) (nullable) (transfer container):
  */
 GList *
 garcon_menu_get_elements (GarconMenu *menu)
@@ -1797,12 +1811,6 @@ garcon_menu_start_monitoring (GarconMenu *menu)
   /* Let only the root menu monitor menu files, merge files/directories and app dirs */
   if (menu->priv->parent == NULL)
     {
-      /* Create the list for merging consecutive file change events */
-      menu->priv->changed_files = NULL;
-
-      /* Reset the idle source for handling file changes */
-      menu->priv->file_changed_idle = 0;
-
       garcon_menu_monitor_menu_files (menu);
 
       garcon_menu_monitor_files (menu, menu->priv->merge_files,
@@ -1849,7 +1857,10 @@ garcon_menu_stop_monitoring (GarconMenu *menu)
 
   /* Stop the idle source for handling file changes from being invoked */
   if (menu->priv->file_changed_idle != 0)
-    g_source_remove (menu->priv->file_changed_idle);
+    {
+      g_source_remove (menu->priv->file_changed_idle);
+      menu->priv->file_changed_idle = 0;
+    }
 
   /* Free the hash table for merging consecutive file change events */
   _garcon_g_slist_free_full (menu->priv->changed_files, g_object_unref);
@@ -1886,7 +1897,7 @@ garcon_menu_monitor_menu_files (GarconMenu *menu)
       /* Build ${XDG_MENU_PREFIX}applications.menu filename */
       prefix = g_getenv ("XDG_MENU_PREFIX");
       relative_filename = g_strconcat ("menus", G_DIR_SEPARATOR_S,
-                                       prefix != NULL ? prefix : GARCON_DEFAULT_MENU_PREFIX,
+                                       prefix != NULL ? prefix : _GARCON_DEFAULT_MENU_PREFIX,
                                        "applications.menu", NULL);
 
       /* Monitor all application menu candidates */
@@ -2143,11 +2154,14 @@ garcon_menu_file_changed (GarconMenu       *menu,
   g_return_if_fail (GARCON_IS_MENU (menu));
   g_return_if_fail (menu->priv->parent == NULL);
 
+  g_mutex_lock (&menu->priv->load_lock);
+
   /* Quick check: reloading is needed if the menu file being used has changed */
   if (g_file_equal (menu->priv->file, file))
     {
       garcon_menu_debug (file, event_type, "menu changed");
       garcon_menu_file_emit_reload_required (menu);
+      g_mutex_unlock (&menu->priv->load_lock);
       return;
     }
 
@@ -2155,12 +2169,15 @@ garcon_menu_file_changed (GarconMenu       *menu,
    * need to refresh if a menu file is removed that we're not using because
    * it is lower in priority (else we'd be using it already) */
   if (event_type == G_FILE_MONITOR_EVENT_DELETED)
-    return;
+    {
+      g_mutex_unlock (&menu->priv->load_lock);
+      return;
+    }
 
   /* Build the ${XDG_MENU_PREFIX}applications.menu filename */
   prefix = g_getenv ("XDG_MENU_PREFIX");
   relative_filename = g_strconcat ("menus", G_DIR_SEPARATOR_S,
-                                   prefix != NULL ? prefix : GARCON_DEFAULT_MENU_PREFIX,
+                                   prefix != NULL ? prefix : _GARCON_DEFAULT_MENU_PREFIX,
                                    "applications.menu", NULL);
 
   /* Get XDG config paths for the root spec (e.g. menus/xfce-applications.menu) */
@@ -2199,6 +2216,8 @@ garcon_menu_file_changed (GarconMenu       *menu,
       garcon_menu_debug (file, event_type, "new menu has higher prio");
       garcon_menu_file_emit_reload_required (menu);
     }
+
+  g_mutex_unlock (&menu->priv->load_lock);
 }
 
 
@@ -2250,6 +2269,8 @@ garcon_menu_process_file_changes (gpointer user_data)
   g_return_val_if_fail (GARCON_IS_MENU (menu), FALSE);
   g_return_val_if_fail (menu->priv->parent == NULL, FALSE);
 
+  g_mutex_lock (&menu->priv->load_lock);
+
   for (lp = menu->priv->changed_files; !stop_processing && lp != NULL; lp = lp->next)
     {
       file = G_FILE (lp->data);
@@ -2337,6 +2358,8 @@ garcon_menu_process_file_changes (gpointer user_data)
   /* reset the idle source ID */
   menu->priv->file_changed_idle = 0;
 
+  g_mutex_unlock (&menu->priv->load_lock);
+
   /* remove the idle source */
   return FALSE;
 }
@@ -2365,6 +2388,8 @@ garcon_menu_app_dir_changed (GarconMenu       *menu,
   g_return_if_fail (GARCON_IS_MENU (menu));
   g_return_if_fail (menu->priv->parent == NULL);
 
+  g_mutex_lock (&menu->priv->load_lock);
+
   if (event_type == G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT
       || event_type == G_FILE_MONITOR_EVENT_CREATED
       || event_type == G_FILE_MONITOR_EVENT_ATTRIBUTE_CHANGED)
@@ -2424,6 +2449,8 @@ garcon_menu_app_dir_changed (GarconMenu       *menu,
             }
         }
     }
+
+  g_mutex_unlock (&menu->priv->load_lock);
 }
 
 
@@ -2439,6 +2466,8 @@ garcon_menu_directory_file_changed (GarconMenu       *menu,
 
   g_return_if_fail (GARCON_IS_MENU (menu));
 
+  g_mutex_lock (&menu->priv->load_lock);
+
   if (event_type == G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT
       || event_type == G_FILE_MONITOR_EVENT_ATTRIBUTE_CHANGED
       || event_type == G_FILE_MONITOR_EVENT_DELETED
@@ -2467,6 +2496,8 @@ garcon_menu_directory_file_changed (GarconMenu       *menu,
       if (old_directory != NULL)
         g_object_unref (old_directory);
     }
+
+  g_mutex_unlock (&menu->priv->load_lock);
 }
 
 
